# Macros for building HPX code.
def hpx_component_copts(component_name):
    """Shorthand for needed HPX component compile options
        Returns the needed HPX component compile options.
    """
    return ['-DHPX_COMPONENT_NAME=' + component_name, '-DHPX_COMPONENT_STRING=\\"' + component_name + '\\"', '-fPIC', '-DHPX_COMPONENT_EXPORTS', '-DHPX_ENABLE_ASSERT_HANDLER']

def hpx_application_copts():
    """Shorthand for needed HPX application compile options
        Returns the needed HPX application compile options.
    """
    return ["-DHPX_APPLICATION_EXPORTS", "-DHPX_ENABLE_ASSERT_HANDLER"]

def hpx_copts():
    """Shorthand for needed HPX compile options
        Returns the needed HPX compile options.
    """
    return "-std=c++17".replace("  ", " ").replace(hpx_path(), "hpx").replace(boost_path(), "boost").split(" ")

def hpx_link_opts():
    """Shorthand for needed HPX link options
        Returns the needed HPX link options.
    """
    return "-std=c++17 -L/usr  -L/usr/lib64 /lib/libhpx_init.a /lib/libhpx.so  -ldl -lrt /usr/lib64/libboost_atomic.so /usr/lib64/libboost_filesystem.so /usr/lib64/libboost_program_options.so /usr/lib64/libboost_regex.so /usr/lib64/libboost_system.so /usr/lib64/libpthread.so /usr/lib64/libboost_atomic.so /usr/lib64/libboost_filesystem.so /usr/lib64/libboost_program_options.so /usr/lib64/libboost_regex.so /usr/lib64/libboost_system.so /usr/lib64/libtcmalloc_minimal.so /usr/lib64/libhwloc.so ${libdir}/libhpx_wrap.a -Wl,-wrap=main".replace("  ", " ").replace(hpx_path(), "hpx").replace(boost_path(), "boost").split(" ")

def hpx_path():
    """Shorthand for the HPX path
        Returns the HPX path.
    """
    return "/usr"

def boost_path():
    """Shorthand for the Boost path
        Returns the Boost path.
    """
    return ""
